from argparse import ArgumentParser
from enum import Enum
from typing import Optional, Sequence

from . import ping
from .probe import probe_mtu


class Mode(Enum):
    """
    Ping compatibility mode

    - auto = auto-detect
    - gnu = GNU ping implementation
    """

    auto = "auto"
    gnu = "gnu"


def positive_int(i) -> int:
    """
    Checks that provided integers are positive and non-null (from CLI args)

    Parameters
    ----------
    i
        String to parse
    """

    x = int(i)

    if x <= 0:
        raise ValueError("Integer must be postitive and non-null")

    return x


def parse_args(argv: Optional[Sequence[str]] = None):
    """
    Setups and executes the CLI arguments parser

    Parameters
    ----------
    argv
        If not specified, argparse will default to the "real" argv
    """

    parser = ArgumentParser(
        description=(
            "mtuprobe uses the system ping utility to determine the IPv4 MTU "
            "between current machine and a network address"
        )
    )

    parser.add_argument(
        "-b", "--ping-bin", default="ping", help="Path to ping binary (default: ping)"
    )
    parser.add_argument(
        "-m",
        "--mode",
        type=Mode,
        default=Mode.auto,
        help=(
            "Compatibility mode of the binary. Options: "
            + ", ".join([x.name for x in Mode])
            + ". (default: auto)"
        ),
    )
    parser.add_argument(
        "-s",
        "--max-size",
        type=positive_int,
        help="Max payload size (default: 3000)",
        default=3000,
    )
    parser.add_argument(
        "-c",
        "--count",
        default=4,
        type=positive_int,
        help="Number of pings sent (default: 4)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=False,
        help="If set, only the MTU will be printed",
    )

    parser.add_argument("address", help="Address you want to ping")

    return parser.parse_args(argv)


def main(argv: Optional[Sequence[str]] = None):
    """
    Executes the main program

    Parameters
    ----------
    argv
        Provide custom CLI args if you call this from Python, otherwise the
        "real ones" will be used
    """

    args = parse_args(argv)
    probe_mtu(
        bin_path=args.ping_bin,
        addr=args.address,
        count=args.count,
        max_size=args.max_size,
        quiet=args.quiet,
        runner=getattr(ping, args.mode.name),
    )


def __main__():
    """
    Wrapper around the main function to catch interrupts
    """

    try:
        main()
    except KeyboardInterrupt:
        pass
    except SystemExit:
        pass


if __name__ == "__main__":
    __main__()



try:
    from collections.abc import Callable  # Python >= 3.7
except ImportError:
    from collections import Callable 


def bisect(n, mapper, tester):
    """
    Runs a bisection.

    - `n` is the number of elements to be bisected
    - `mapper` is a callable that will transform an integer from "0" to "n"
      into a value that can be tested
    - `tester` returns true if the value is within the "right" range
    """

    if n < 1:
        raise ValueError("Cannot bisect an empty array")

    left = 0
    right = n - 1

    while left + 1 < right:
        mid = int((left + right) / 2)

        val = mapper(mid)

        if tester(val):
            right = mid
        else:
            left = mid

    return mapper(right)


def probe_mtu(
    bin_path: str, addr: str, count: int, max_size: int, quiet: bool, runner: Callable
) -> None:
    """
    Uses a bisection algorithm to pinpoint the exact MTU of the path

    Parameters
    ----------
    bin_path
        Path to the ping binary
    addr
        Address to ping
    count
        Number of pings to do for each try
    max_size
        Max MTU size to test
    quiet
        Stay quiet in the output (just print the found MTU)
    runner
        Ping runner function
    """

    def mapper(n):
        return n + 1

    def tester(n):
        if not quiet:
            print(f"\rTesting packet size: {n: >6}", end="", flush=True)

        return not runner(
            bin_path=bin_path,
            addr=addr,
            count=count,
            size=n,
        )

    size = bisect(max_size, mapper, tester) - 1

    if not quiet:
        extra = 8 + 20  # ICMP + IPv4

        print(f"\rMax packet size:       {size: >6}")
        print(f"Expected ethernet MTU: {size + extra: >6}")
    else:
        print(f"{size}")



from subprocess import DEVNULL, Popen
from sys import stderr


def auto(bin_path: str, addr: str, count: int, size: int) -> bool:
    """
    In the future, auto-detect which ping implementation to call, however for
    now just forwards the call to gnu.

    Returns True if at least one ping went through (and came back)

    Parameters
    ----------
    bin_path
        Path to ping binary
    addr
        Address to ping
    count
        Number of pings to send
    size
        Size of the payload
    """

    return gnu(bin_path, addr, count, size)


def gnu(bin_path: str, addr: str, count: int, size: int) -> bool:
    """
    Using the GNU ping (so Linux machines most likely)
    """

    try:
        with Popen(
            args=[
                bin_path,
                "-i",
                "0.2",
                "-c",
                f"{count}",
                "-M",
                "do",
                "-s",
                f"{size}",
                addr,
            ],
            stdout=DEVNULL,
            stdin=DEVNULL,
            stderr=DEVNULL,
        ) as p:
            p.wait()
    except FileNotFoundError:
        stderr.write(f'\rping binary "{bin_path}" could not be found\n')
        stderr.flush()
        exit(1)
    except PermissionError:
        stderr.write(f'\rping binary "{bin_path}" cannot be executed\n')
        stderr.flush()
        exit(1)

    return not p.returncode
